
//package com.knownspace.tools;

import MalformedDelimiterException;

import com.knownspace.tools.StringSearcher;
import com.knownspace.tools.StringTools;
import com.knownspace.tools.MalformedPatternException;

import java.io.IOException;

public class ParenthesizedExtractor
	{
	/*
	Scan a string and extract text parenthesized by arbitrary strings.
	The delimiters themselves are not returned.
	After each extraction, the current character is the first character
	occurring after the last character of the last delimiter searched for.

	Public Methods and Constructors:
	--------------------------------
	ParenthesizedExtractor(String targetString, String leftDelimiter,
		String rightDelimiter, boolean collapseWhitespace, boolean ignoreCase)
		initialize a new ParenthesizedExtractor to use the given delimiters
		with the given target string, perhaps collapse each run of whitespace to a
		single space, and perhaps ignore the case of characters during searches

	ParenthesizedExtractor(String targetString, String leftDelimiter,
		String rightDelimiter, boolean collapseWhitespace)
		initialize a new ParenthesizedExtractor to use the given delimiters
		with the given target string, perhaps collapse each run of whitespace
		to a single space, and ignore the case of characters during searches

	ParenthesizedExtractor(String targetString, String leftDelimiter,
		String rightDelimiter)
		initialize a new ParenthesizedExtractor to use the given delimiters
		with the given target string, collapse each run of whitespace to a
		single space, and ignore the case of characters during searches

	setDelimiters(String leftDelimiter, String rightDelimiter)
		reset the delimiters to be used in subsequent searches

	skipForward(numCharacters)
		reset the search position to numCharacters after the current position

	skipBackward(numCharacters)
		reset the search position to numCharacters before the current position

	reset()
		reset the search position to the start of the text stream

	String getNext(String leftDelimiter, String rightDelimiter)
		return all text found between the next occurrence of the given
		delimiters; return null if either of the delimiters is not found

	String getNext()
		return all text found between the next occurrence of the preset
		delimiters; return null if either of the delimiters is not found

	String getUntil(String delimiter)
		return all text found from the present position to the next
		occurrence of delimiter; return null if the delimiter is not found

	String getAfter(String delimiter)
		return all text found after the next occurrence of delimiter
		until the end of the string; return null if the delimiter is not found

	Class Relationships:
	--------------------
	This class uses objects of class StringSearcher to do the actual searching.
	This class can throw objects of class MalformedDelimiterException.
	*/

	//the text stream to scan for delimited text
	private String targetString;

	//the uppercase version of the text stream
	//(to be used only if ignoreCase is set)
	private String targetStringUppercase;

	//the index of the current character in the text stream
	private int targetCursor;

	//the current default delimiters to use
	private String leftDelimiter, rightDelimiter;

	//collapse runs of whitespace in the text stream?
	private boolean collapseWhitespace;

	//ignore the case of characters during searches?
	private boolean ignoreCase;

	/////////////////////////////////////////////////////////////////////
	public ParenthesizedExtractor(final String textString,
		final String leftDelimiter, final String rightDelimiter)
		throws MalformedDelimiterException
		{
		/*
		Initialize a new ParenthesizedExtractor to read from a string with
		given delimiters, collapse all whitespace runs to single spaces,
		and ignore the case of characters during searches.

		Throw a MalformedDelimiterException if either of the two delimiters
		is null or the empty string.
		*/

		this(textString, leftDelimiter, rightDelimiter, true, true);
		}

	/////////////////////////////////////////////////////////////////////
	public ParenthesizedExtractor(final String textString,
		final String leftDelimiter, final String rightDelimiter,
		final boolean collapseWhitespace)
		throws MalformedDelimiterException
		{
		/*
		Initialize a new ParenthesizedExtractor to read from a string with
		given delimiters, perhaps collapse all whitespace runs to single
		spaces, and ignore the case of characters during searches.

		Throw a MalformedDelimiterException if either of the two delimiters
		is null or the empty string.
		*/

		this(textString, leftDelimiter, rightDelimiter, collapseWhitespace, true);
		}

	/////////////////////////////////////////////////////////////////////
	public ParenthesizedExtractor(final String textString,
		final String leftDelimiter, final String rightDelimiter,
		final boolean collapseWhitespace, final boolean ignoreCase)
		throws MalformedDelimiterException
		{
		/*
		Initialize a new ParenthesizedExtractor to read from a stream with
		given delimiters. Perhaps collapse all whitespace runs in the text
		stream to single spaces. Perhaps ignore the case of characters during
		searches.

		Throw a MalformedDelimiterException if either of the two delimiters
		is null or the empty string.
		*/

		this.ignoreCase = ignoreCase;
		this.collapseWhitespace = collapseWhitespace;
		saveStreamToString(textString);
		targetCursor = 0;
		setDelimiters(leftDelimiter, rightDelimiter);
		}

	/////////////////////////////////////////////////////////////////////
	private void saveStreamToString(final String inputString)
		{
		/*
		Create a String containing all the characters in inputString.

		If ignoreCase is set, create an uppercase version of the inputString.

		If collapseWhitespace is set, collapse each run of whitespace into
		one space character unless the run occurs inside a quote. Eat any
		leading whitespaces as well.
		*/

		StringBuffer inputStreamString = new StringBuffer("");
		boolean insideQuote = false;
		int thisCharacter = inputString.charAt(0);
		int lastCharacter = ' ';
		for (int i = 0; i < inputString.length(); i++)
			{
			//keep track of whether i'm presently inside a quote
			//(don't be fooled by an escaped quote character)
			if ((thisCharacter == '\"') && (lastCharacter != '\\'))
				insideQuote = !insideQuote;

			//if collapseWhitespace is set and i'm not inside a quote
			//make all whitespaces into simple spaces
			if (collapseWhitespace && !insideQuote &&
				Character.isWhitespace((char) thisCharacter))
				thisCharacter = ' ';

			//add the current character unless it's a space, the last one was
			//a space, collapseWhitespace is set, and i'm not inside a quote
			if (!collapseWhitespace || insideQuote ||
				(thisCharacter != ' ') || (lastCharacter != ' '))
				inputStreamString.append((char) thisCharacter);

			lastCharacter = thisCharacter;
			thisCharacter = inputString.charAt(i);
			}

		//save the text stream and perhaps create an uppercase version
		targetString = inputStreamString.toString();
		if (ignoreCase)
			targetStringUppercase = targetString.toUpperCase();
		}

	/////////////////////////////////////////////////////////////////////
	public void setDelimiters(final String leftDelimiter,
		final String rightDelimiter)
		throws MalformedDelimiterException
		{
		/*
		Set the delimiters to new values.

		Throw a MalformedDelimiterException if either delimiter is null or
		the empty string.
		*/

		checkDelimiter(leftDelimiter);
		checkDelimiter(rightDelimiter);

		this.leftDelimiter = leftDelimiter;
		this.rightDelimiter = rightDelimiter;
		}

	/////////////////////////////////////////////////////////////////////
	private void checkDelimiter(final String delimiter)
		throws MalformedDelimiterException
		{
		/*
		Check whether a delimiter has a legal value.

		Throw a MalformedDelimiterException if the delimiter is null or the
		empty string.
		*/

		if ((delimiter == null) || (delimiter == ""))
			throw new MalformedDelimiterException();
		}

	/////////////////////////////////////////////////////////////////////
	public void reset()
		{
		/*
		Reset the current character position so that subsequent searches
		begin at the beginning of the text stream.
		*/

		targetCursor = 0;
		}

	/////////////////////////////////////////////////////////////////////
	public void skipForward(final int numCharacters)
		{
		/*
		Skip over the next numCharacters characters of the text stream.
		*/

		targetCursor += numCharacters;
		if (targetCursor > targetString.length() - 1)
			targetCursor = targetString.length() - 1;
		}

	/////////////////////////////////////////////////////////////////////
	public void skipBackward(final int numCharacters)
		{
		/*
		Skip back the last numCharacters characters of the text stream.
		*/

		targetCursor -= numCharacters;
		if (targetCursor < 0)
			targetCursor = 0;
		}

	/////////////////////////////////////////////////////////////////////
	private StringSearcher setSearcher(final String delimiter)
		throws MalformedDelimiterException
		{
		/*
		Return a StringSearcher with the delimiter as the search pattern.

		Throw a MalformedDelimiterException if the delimiter is null or the
		empty string.
		*/

		checkDelimiter(delimiter);

		StringSearcher searcher = null;

		try
			{
			searcher = new StringSearcher(delimiter);
			}
		catch (MalformedPatternException ignored)
			{ //should never happen!
			System.out.println("internal error");
			ignored.printStackTrace();
			System.exit(1);
			}

		return searcher;
		}

	/////////////////////////////////////////////////////////////////////
	private int searchTargetString(final String delimiter)
		throws MalformedDelimiterException
		{
		/*
		Create a string searcher for the delimiter and search the text stream
		using it. If the delimiter is found, return the character position of
		the first character in the text stream where delimiter next occurs,
		otherwise return StringSearcher.PATTERN_NOT_FOUND.

		If ignoreCase is set, search the uppercase version of the text stream,
		else search the normal text stream.

		Future optimization: keep track of the StringSearchers created so that
		existing ones need not be recreated.

		Throw a MalformedDelimiterException if the delimiter is null or the
		empty string.
		*/

		StringSearcher searcher;
		if (ignoreCase)
			{
			searcher = setSearcher(delimiter.toUpperCase());
			return searcher.find(targetStringUppercase, targetCursor);
			}
		else
			{
			searcher = setSearcher(delimiter);
			return searcher.find(targetString, targetCursor);
			}
		}

	/////////////////////////////////////////////////////////////////////
	public String getNext()
		throws MalformedDelimiterException
		{
		/*
		Return all text after the next occurrence of the preset leftDelimiter
		until, but not including, the subsequent occurrence of the preset
		rightDelimiter.

		If either delimiter is not found, leave the current character position
		unchanged and return null.
		If both delimiters are found, make the current character the first
		character after the last character of the next occurrence of
		rightDelimiter in the text stream.

		Throw a MalformedDelimiterException if either delimiter is null or the
		empty string. (Note: this should be optimized away since the delimiters
		are preset then they have already been checked.)
		*/

		return getNext(leftDelimiter, rightDelimiter);
		}

	/////////////////////////////////////////////////////////////////////
	public String getNext(final String leftDelimiter,
		final String rightDelimiter)
		throws MalformedDelimiterException
		{
		/*
		Return all text after the next occurrence of leftDelimiter until,
		but not including, the subsequent occurrence of rightDelimiter.

		If either delimiter is not found, leave the current character position
		unchanged and return null.
		If both delimiters are found, make the current character the first
		character after the last character of the next occurrence of
		rightDelimiter in the text stream.

		Throw a MalformedDelimiterException if either delimiter is null or the
		empty string.
		*/

		//search for leftDelimiter in the text stream
		int found = searchTargetString(leftDelimiter);

		//return null if leftDelimiter doesn't occur in the remaining text
		//stream
		String delimitedText = null;
		if (found != StringSearcher.PATTERN_NOT_FOUND)
			{
			//found points to the first character of leftDelimiter
			//in the text stream
			targetCursor = found + leftDelimiter.length();
			delimitedText = getUntil(rightDelimiter);
			}

		return delimitedText;
		}

	/////////////////////////////////////////////////////////////////////
	public String getUntil(final String delimiter)
		throws MalformedDelimiterException
		{
		/*
		Return all text from the current character until, but not including,
		the first character of the next occurrence of delimiter.

		If the delimiter is not found, leave the current character position
		unchanged and return null.
		If the delimiter is found, make the current character the first
		character after the last character of the next occurrence of delimiter
		in the text stream.

		Throw a MalformedDelimiterException if the delimiter is null or the
		empty string.
		*/

		//search for the delimiter in the text stream
		int found = searchTargetString(delimiter);

		//return null if delimiter doesn't occur in the remaining text stream
		String delimitedText = null;
		if (found != StringSearcher.PATTERN_NOT_FOUND)
			{
			//found points to the first character of delimiter in the text stream
			delimitedText = reconstructString(targetCursor, found - 1);
			targetCursor = found + delimiter.length();
			}

		return delimitedText;
		}

	/////////////////////////////////////////////////////////////////////
	public String getAfter(final String delimiter)
		throws MalformedDelimiterException
		{
		/*
		Return all text after the last character of the next occurrence
		of delimiter until the end of the text stream.

		If the delimiter is not found, leave the current character position
		unchanged and return null.
		If the delimiter is found, make the current character the first
		character after the last character of the next occurrence of delimiter
		in the text stream.

		Throw a MalformedDelimiterException if the delimiter is null or the
		empty string.
		*/

		//search for the delimiter in the text stream
		int found = searchTargetString(delimiter);

		//return null if delimiter doesn't occur in the remaining text stream;
		String delimitedText = null;
		if (found != StringSearcher.PATTERN_NOT_FOUND)
			{
			//found points to the first character of delimiter in the text stream
			delimitedText = reconstructString(found + delimiter.length(),
				targetString.length() - 1);
			targetCursor = found + delimiter.length();
			}

		return delimitedText;
		}

	/////////////////////////////////////////////////////////////////////
	private String reconstructString(final int lowerIndex,
		final int upperIndex)
		{
		/*
		Return a substring of the text stream from lowerIndex to upperIndex
		that also includes any previously subtracted strings.
		*/

		StringBuffer reconstructedString = new StringBuffer("");
		for (int i = lowerIndex; i <= upperIndex; i++)
			{
			reconstructedString.append(targetString.charAt(i));
			}

		return reconstructedString.toString();
		}

	/////////////////////////////////////////////////////////////////////
	public static void main(final String[] commandLineArguments)
		throws IllegalArgumentException, MalformedDelimiterException, IOException
		{
		/*
		Exercise the above code.
		*/

		if (commandLineArguments.length != 1)
			throw new IllegalArgumentException("Usage: filename");

		String string = StringTools.convertFileToString(commandLineArguments[0]);

		ParenthesizedExtractor extractor =
			new ParenthesizedExtractor(string, "<", ">", true, true);

		System.out.println("TEST OF getAfter(\"THE\")");
		String extractedText = extractor.getAfter("THE");
		while (extractedText != null)
			{
			System.out.println("extracted: " + extractedText);
			extractedText = extractor.getAfter("THE");
			}

		System.out.println();
		System.out.println("TEST OF getUntil(\">\")");
		extractor.reset();
		extractedText = extractor.getUntil(">");
		while (extractedText != null)
			{
			System.out.println("extracted: " + extractedText);
			extractedText = extractor.getUntil(">");
			}

		System.out.println();
		System.out.println("TEST OF getNext(\"<\", \">\")");
		extractor.reset();
		extractedText = extractor.getNext();
		while (extractedText != null)
			{
			System.out.println("extracted: " + extractedText);
			extractedText = extractor.getNext();
			}
		}
	}
