
package com.knownspace.tools;

import java.util.Vector;
import java.util.Enumeration;

public class Clock
	implements Runnable
	{
	/*
	Define an object that regularly executes the tick() method of all
	its observers.

	Objects become observers by executing this clock's addClockObserver()
	method or by passing themselves to this clock on its creation.

	Public Constructors & Methods:
	------------------------------
	Clock()
	Clock(ticksPerSecond)
	Clock(ticksPerSecond, observer)
	addClockObserver(observer)
	deleteClockObserver(observer)
	start()
	stop()
	run()
	timeSinceLastTick()
	*/

	//my thread, observer list, clock rate, and temporary variables
	protected Thread ticker;
	protected Vector observers = new Vector();
	protected int ticksPerSecond = 25;
	protected long sleepTime = 40;
	protected long startTickTime= 0, currentTickTime= 0, lastTickTime= 0;
	protected long tickCount = 0;

	///////////////////////////////////////////////////////////////////////
	public Clock()
		{
		/*
		Create a clock with a default speed and no observers.
		*/

		ticker = new Thread(this);
		ticker.setDaemon(true);
		}

	///////////////////////////////////////////////////////////////////////
	public Clock(int ticksPerSecond)
		{
		/*
		Create a clock with a given speed and no observers.
		*/

		ticker = new Thread(this);
		ticker.setDaemon(true);
		this.ticksPerSecond = ticksPerSecond;
		//convert to milliseconds
		sleepTime = (long) (1000 / ticksPerSecond);
		}

	///////////////////////////////////////////////////////////////////////
	public Clock(int ticksPerSecond, ClockObserver observer)
		{
		/*
		Create a clock with a given speed and a given observer
		and start it.
		*/

		ticker = new Thread(this);
		ticker.setDaemon(true);
		this.ticksPerSecond = ticksPerSecond;
		//convert to milliseconds
		sleepTime = (long) (1000 / ticksPerSecond);
		addClockObserver(observer);
		this.start();
		}

	///////////////////////////////////////////////////////////////////////
	public synchronized void addClockObserver(ClockObserver observer)
		{
		observers.addElement(observer);
		}

	///////////////////////////////////////////////////////////////////////
	public synchronized void deleteClockObserver(ClockObserver observer)
		{
		observers.removeElement(observer);
		}

	///////////////////////////////////////////////////////////////////////
	public synchronized void die(ClockObserver observer)
		{
		/*
		Die if requested to by my last observer.

		Note: this would leave the observer (and any careless past observers)
		with a null pointer. how to fix that without requiring them to remember
		to null out their references to me? need some way to make myself private
		while still allowing arbitrary objects to observe me. I need a
		ClockManager.
		*/

		if ((observers.size() == 1) && observers.elementAt(0).equals(observer))
			{
			this.stop();
			observers = null;
			try {this.finalize();}
			catch (Throwable ignored) {} //just die
			}
		}

	///////////////////////////////////////////////////////////////////////
	public void start()
		{
		startTickTime = System.currentTimeMillis();
		lastTickTime = startTickTime;
		currentTickTime = startTickTime;

		if (ticker == null)
			{
			ticker = new Thread(this);
			ticker.setDaemon(true);
			}
		ticker.start();
		}

	///////////////////////////////////////////////////////////////////////
	public void stop()
		{
		ticker = null;
		}

	///////////////////////////////////////////////////////////////////////
	public void run()
		{
		while (ticker != null)
			{
			currentTickTime = System.currentTimeMillis();

			//execute all my observers' tick() methods
			Enumeration watchers = observers.elements();
			while (watchers.hasMoreElements())
				((ClockObserver) watchers.nextElement()).tick();
			tickCount++;

			long elapsedTime = System.currentTimeMillis() - currentTickTime;

			lastTickTime = currentTickTime;

			//the time to sleep depends on how long it took to call all the
			//tick() methods, so sleep only long enough to keep the beat
			if (elapsedTime < sleepTime)
				{
				try {Thread.sleep(sleepTime - elapsedTime);}
				catch (InterruptedException exception)
					{
					System.out.println("Clock: something bad happened.");
					exception.printStackTrace();
					throw new RuntimeException(exception.toString());
					}
				}
			}
		}

	///////////////////////////////////////////////////////////////////////
	public long timeSinceLastTick()
		{
		return currentTickTime - lastTickTime;
		}

	///////////////////////////////////////////////////////////////////////
	protected long getTickCount()
		{
		return tickCount;
		}
	}
