
package com.knownspace.tools;

import java.io.CharArrayWriter;
import java.io.PrintWriter;
import java.util.StringTokenizer;
import java.util.NoSuchElementException;

public class Debug
	{
	/*
	A simple debugger to implement breakpoint assertion testing.

	Public Methods:
	---------------
	assert(boolean)         -- fires if condition is false
	assert(boolean, string) -- fires if condition is false and prints string

	In both cases the debugger prints the offending method's name.

	Every method should have its own assertions to test the state of its
	arguments on entry into the method.
	*/

	private final static boolean FATAL_ERROR_LEVEL_SET = false;

	///////////////////////////////////////////////////////////////////////
	public final static void assert(boolean condition)
		{
		if (! condition)
			printError("assertion failed in: " + getCaller(2));

		checkFatality();
		}

	///////////////////////////////////////////////////////////////////////
	public final static void assert(boolean condition, String string)
		{
		if (! condition)
			{
			printError("assertion failed: " + string);
			printError("in: " + getCaller(2));
			}

		checkFatality();
		}

	///////////////////////////////////////////////////////////////////////
	private final static void printError(String string)
		{
		System.out.println(string);
		}

	///////////////////////////////////////////////////////////////////////
	private final static void checkFatality()
		{
		if (FATAL_ERROR_LEVEL_SET)
			System.exit(1);
		}

	///////////////////////////////////////////////////////////////////////
	private final static String getCaller(int callerNumber)
		{
		/*
		Determine the callerNumber'th method that called this method
		and return its name.

		NOTE: This method may not work with code compiled using a JIT
		or with a non-JDK JVM since in either case the stack trace
		format is unspecified in the language spec.

		Stack trace format assumed:
		---------------------------
		java.lang.Throwable
			at <this method's name>
			at <caller's method's name>                   (callerNumber == 1)
			at <caller's caller's method's name>          (callerNumber == 2)
			at <caller's caller's caller's method's name> (callerNumber == 3)
		*/

		if (callerNumber < 1)
			return "Debugger";

		//transform the caller number into the token number to search for
		int tokenNumber = (callerNumber * 2) + 3;

		//create a stream to catch the stack trace
		CharArrayWriter stream = new CharArrayWriter();

		//send the stack trace to the stream
		Throwable throwable = new Throwable();
		throwable.printStackTrace(new PrintWriter(stream));

		//parse the stack trace to find the callerNumber'th method call
		StringTokenizer stackTrace = new StringTokenizer(stream.toString());
		String token = "";
		try
			{
			for (int i = 0; i < tokenNumber; i++)
				token = stackTrace.nextToken();
			}
		catch (NoSuchElementException ignored)
			{
			token = "unknown";
			}

		return token;
		}
	}
