// Modified from _Java Examples in a Nutshell_. (http://www.oreilly.com)
// Copyright (c) 1997 by David Flanagan

import java.io.IOException;
import java.io.File;
import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.FileInputStream;
import java.io.FileOutputStream;

public class FileCopy
	{
	/*
	This class is a standalone program to copy a file, and also defines a 
	static copy() method that other programs can use to copy files.
	*/

	////////////////////////////////////////////////////////////////////////
	public static void main(String[] commandLineArguments)
		{
		/*
		The main() method of the standalone program.
		*/

		if (commandLineArguments.length != 2)
			System.err.println("Usage: java FileCopy <source file> <destination>");
		else
			{
			try
				{
				copy(commandLineArguments[0], commandLineArguments[1]);
				}
			catch (IOException exception)
				{
				System.err.println(exception.getMessage());
				}
			}
		}

	////////////////////////////////////////////////////////////////////////
	private static void abort(String message)
		throws IOException
		{
		/*
		A convenience method to throw an exception
		*/

		throw new IOException(message);
		}

	////////////////////////////////////////////////////////////////////////
	public static void copy(String fromFilename, String toFilename)
		throws IOException
		{
		/*
		Make sure that file fromFilename can be copied to destination toFilename
		then copy it.
		*/

		//convert the filename strings to file objects
		File fromFile = new File(fromFilename);
		File toFile = new File(toFilename);

		//make sure the source file exists, is a file, and is readable
		if (!fromFile.exists())
			abort("FileCopy: unable to find: " + fromFilename);
		if (!fromFile.isFile())
			abort("FileCopy: can't copy directory: " + fromFilename);
		if (!fromFile.canRead())
			abort("FileCopy: " + fromFilename + " is unreadable");

		//if the destination is a directory, use the source file name
		//as the destination file name
		if (toFile.isDirectory())
			toFile = new File(toFile, fromFile.getName());

		//if the destination file exists,
		//make sure it's writeable and ask before overwriting it
		if (toFile.exists())
			{
			if (!toFile.canWrite())
				abort("FileCopy: destination file is unwriteable: " + toFilename);

			//overwrite it?
			System.out.print("Overwrite file " + toFilename + "?" +
				"(please respond yes or no): ");
			System.out.flush();
			BufferedReader in =
				new BufferedReader(new InputStreamReader(System.in));
			String response = in.readLine();
			if (! response.toUpperCase().startsWith("Y"))
				abort("FileCopy: existing file was not overwritten.");
			}

		//if the destination file doesn't exist,
		//check if its directory exists and is writeable;
		//if its parent doesn't exist either, use the current directory
		if (!toFile.exists())
			{
			String parent = toFile.getParent();
			if (parent == null)
				parent = System.getProperty("user.dir");
			File dir = new File(parent);
			if (!dir.exists())
				abort("FileCopy: destination directory doesn't exist: " + parent);
			if (dir.isFile())
				abort("FileCopy: destination is not a directory: " + parent);
			if (!dir.canWrite())
				abort("FileCopy: destination directory is unwriteable: " + parent);
			}

		//okay, everything checked; copy the file
		copyFile(fromFile, toFile);
		}

	////////////////////////////////////////////////////////////////////////
	public static void copyFile(File fromFile, File toFile)
		throws IOException
		{

		// Stream to read from source
		FileInputStream from = null;

		// Stream to write to destination
		FileOutputStream to = null;

		try
			{
			//create input and output streams
			from = new FileInputStream(fromFile);
			to = new FileOutputStream(toFile);

			//a buffer to hold file contents
			byte[] buffer = new byte[4096];
			//how many bytes in buffer
			int bytesRead;

			//read a chunk of bytes into the buffer, then write them out,
			//until end of the file (when read() returns -1).
			while((bytesRead = from.read(buffer)) != -1)
				to.write(buffer, 0, bytesRead);
			}

		//close the streams, even if exceptions were thrown
		finally
			{
			if (from != null)
				{
				try { from.close(); }
				catch (IOException e) { ; }
				}
			if (to != null)
				{
				try { to.close(); }
				catch (IOException e) { ; }
				}
			}
		}
	}
